//=============================================================================
// Camera.cpp by Frank Luna (C) 2004 All Rights Reserved.
//=============================================================================

#include "Camera.h"
//#include "d3dUtility.h"

Camera* gCamera = 0;

Camera::Camera()
{
	mCameraType = AIRCRAFT;

	D3DXMatrixIdentity(&mView);

	mPosW   = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	mRightW = D3DXVECTOR3(1.0f, 0.0f, 0.0f);
	mUpW    = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	mLookW  = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
}

Camera::Camera(CameraType camType)
{
	mCameraType = camType;

	D3DXMatrixIdentity(&mView);

	mPosW   = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	mRightW = D3DXVECTOR3(1.0f, 0.0f, 0.0f);
	mUpW    = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	mLookW  = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
}

void Camera::setCameraType(CameraType cameraType)
{
	mCameraType = cameraType;
}

void Camera::setPosition(D3DXVECTOR3 pos)
{
	mPosW = pos;
}

D3DXVECTOR3& Camera::pos()
{
	return mPosW;
}

const D3DXVECTOR3& Camera::right() const
{
	return mRightW;
}

const D3DXVECTOR3& Camera::up() const
{
	return mUpW;
}

const D3DXVECTOR3& Camera::look() const
{
	return mLookW;
}

const D3DXMATRIX& Camera::view() const
{
	return mView;
}

void Camera::walk(float units)
{
	// move only on xz plane for land object
	if( mCameraType == LANDOBJECT )
		mPosW += D3DXVECTOR3(mLookW.x, 0.0f, mLookW.z) * units;

	if( mCameraType == AIRCRAFT )
		mPosW += mLookW * units;
}

void Camera::strafe(float units)
{
	// move only on xz plane for land object
	if( mCameraType == LANDOBJECT )
		mPosW += D3DXVECTOR3(mRightW.x, 0.0f, mRightW.z) * units;

	if( mCameraType == AIRCRAFT )
		mPosW += mRightW * units;
}

void Camera::fly(float units)
{
	// move only on y-axis for land object
	if( mCameraType == LANDOBJECT )
		mPosW.y += units;

	if( mCameraType == AIRCRAFT )
		mPosW += mUpW * units;
}

void Camera::pitch(float angle)
{
	D3DXMATRIX R;
	D3DXMatrixRotationAxis(&R, &mRightW, angle);

	// rotate _up and _look around _right vector
	D3DXVec3TransformCoord(&mUpW, &mUpW, &R);
	D3DXVec3TransformCoord(&mLookW, &mLookW, &R);
}

void Camera::yaw(float angle)
{
	D3DXMATRIX R;

	// rotate around world y (0, 1, 0) always for land object
	if( mCameraType == LANDOBJECT )
		D3DXMatrixRotationY(&R, angle);

	// rotate around own up vector for aircraft
	if( mCameraType == AIRCRAFT )
		D3DXMatrixRotationAxis(&R, &mUpW, angle);

	// rotate _right and _look around _up or y-axis
	D3DXVec3TransformCoord(&mRightW,&mRightW, &R);
	D3DXVec3TransformCoord(&mLookW,&mLookW, &R);
}

void Camera::roll(float angle)
{
	// only roll for aircraft type
	if( mCameraType == AIRCRAFT )
	{
		D3DXMATRIX R;
		D3DXMatrixRotationAxis(&R, &mLookW,	angle);

		// rotate _up and _right around _look vector
		D3DXVec3TransformCoord(&mRightW,&mRightW, &R);
		D3DXVec3TransformCoord(&mUpW,&mUpW, &R);
	}
}

void Camera::update()
{
	// Keep camera's axes orthogonal to eachother
	D3DXVec3Normalize(&mLookW, &mLookW);

	D3DXVec3Cross(&mUpW, &mLookW, &mRightW);
	D3DXVec3Normalize(&mUpW, &mUpW);

	D3DXVec3Cross(&mRightW, &mUpW, &mLookW);
	D3DXVec3Normalize(&mRightW, &mRightW);

	// Build the view matrix:
	float x = -D3DXVec3Dot(&mRightW, &mPosW);
	float y = -D3DXVec3Dot(&mUpW, &mPosW);
	float z = -D3DXVec3Dot(&mLookW, &mPosW);

	mView(0,0) = mRightW.x; mView(0, 1) = mUpW.x; mView(0, 2) = mLookW.x; mView(0, 3) = 0.0f;
	mView(1,0) = mRightW.y; mView(1, 1) = mUpW.y; mView(1, 2) = mLookW.y; mView(1, 3) = 0.0f;
	mView(2,0) = mRightW.z; mView(2, 1) = mUpW.z; mView(2, 2) = mLookW.z; mView(2, 3) = 0.0f;
	mView(3,0) = x;        mView(3, 1) = y;     mView(3, 2) = z;       mView(3, 3) = 1.0f;
}

D3DXMATRIX& Camera::lookAt(D3DXVECTOR3& pos, D3DXVECTOR3& target, D3DXVECTOR3& up)
{
	D3DXVECTOR3 L = target - pos;
	D3DXVec3Normalize(&L, &L);

	D3DXVECTOR3 R;
	D3DXVec3Cross(&R, &up, &L);
	D3DXVec3Normalize(&R, &R);

	D3DXVECTOR3 U;
	D3DXVec3Cross(&U, &L, &R);
	D3DXVec3Normalize(&U, &U);

	mPosW   = pos;
	mRightW = R;
	mUpW    = U;
	mLookW  = L;

	// Keep camera's axes orthogonal to each other and of unit length.
	D3DXVec3Normalize(&mLookW, &mLookW);

	D3DXVec3Cross(&mUpW, &mLookW, &mRightW);
	D3DXVec3Normalize(&mUpW, &mUpW);

	D3DXVec3Cross(&mRightW, &mUpW, &mLookW);
	D3DXVec3Normalize(&mRightW, &mRightW);

	// Fill in the view matrix entries.
	float x = -D3DXVec3Dot(&mPosW, &mRightW);
	float y = -D3DXVec3Dot(&mPosW, &mUpW);
	float z = -D3DXVec3Dot(&mPosW, &mLookW);

	mView(0,0) = mRightW.x; 
	mView(1,0) = mRightW.y; 
	mView(2,0) = mRightW.z; 
	mView(3,0) = x;   

	mView(0,1) = mUpW.x;
	mView(1,1) = mUpW.y;
	mView(2,1) = mUpW.z;
	mView(3,1) = y;  

	mView(0,2) = mLookW.x; 
	mView(1,2) = mLookW.y; 
	mView(2,2) = mLookW.z; 
	mView(3,2) = z;   

	mView(0,3) = 0.0f;
	mView(1,3) = 0.0f;
	mView(2,3) = 0.0f;
	mView(3,3) = 1.0f;

	return mView;
}
